#include <allegro.h>
#include <aldumb.h>
#include "pmask/pmask.h"
#include "ticker.h"
#include "music.h"

#include "truck.h"

int sound = 0;
int user_exit = 0;

BITMAP * back_buffer = NULL;
BITMAP * top_buffer = NULL;
BITMAP * bottom_buffer = NULL;

/* game variables */
int state = 0;
int game_state = 0;
int train_pos;
BITMAP * train_pic = NULL;
PALETTE game_palette;

TRUCK truck;
/*BITMAP * truck_pic = NULL;
fixed truck_x, truck_y;
unsigned char truck_angle;
BITMAP * truck_mask_pic = NULL;
PMASK * truck_mask; */

BITMAP * level_pic = NULL;
int level_x, level_y;
BITMAP * level_mask_pic = NULL;
PMASK * level_mask = NULL;
BITMAP * piece_pic[9] = {NULL};
char piece_slot[5] = {0, 1, 2, 3, 4};
BITMAP * track_pic = NULL;
BITMAP * game_over_pic = NULL;

char piece[4][4];
char board[6][20];

void generate_board(void)
{
	int i, j;
	
	for(i = 0; i < 6; i++)
	{
		for(j = 0; j < 20; j++)
		{
			if(rand() % 1000 < 250)
			{
				board[i][j] = 1;
			}
			else
			{
				board[i][j] = 0;
			}
		}
	}
}

void draw_ascii(BITMAP * dest, BITMAP * src)
{
	int i, j;
	
	for(i = 0; i < src->h; i++)
	{
		for(j = 0; j < src->w; j++)
		{
			if(src->line[i][j])
			{
				textprintf_ex(dest, font, j * 8, i * 8, makecol(255, 0, 0), makecol(0, 0, 0), "+");
			}
			else
			{
				textprintf_ex(dest, font, j * 8, i * 8, makecol(0, 0, 255), makecol(0, 0, 0), "-");
			}
		}
	}
}

int tins06_initialize(void)
{
	if(allegro_init())
	{
		allegro_message("Allegro failed to initialize!");
		return 0;
	}
	if(set_gfx_mode(GFX_AUTODETECT_WINDOWED, 640, 480, 0, 0))
	{
		if(set_gfx_mode(GFX_AUTODETECT, 640, 480, 0, 0))
		{
			allegro_message("Could not set graphics mode!");
			return 0;
		}
	}
	back_buffer = create_bitmap(640, 480);
	if(!back_buffer)
	{
		allegro_message("Could not create backbuffer!");
		return 0;
	}
	top_buffer = create_sub_bitmap(back_buffer, 0, 0, 640, 240);
	bottom_buffer = create_sub_bitmap(back_buffer, 0, 240, 640, 240);
	if(install_keyboard())
	{
		allegro_message("Could not initialize keyboard!");
		return 0;
	}
	if(install_mouse() < 0)
	{
		allegro_message("Could not initialize mouse!");
		return 0;
	}
	if(install_sound(DIGI_AUTODETECT, MIDI_NONE, NULL))
	{
		allegro_message("Could not initialize sound driver!");
		return 0;
	}
	if(install_timer())
	{
		allegro_message("Could not initialize timer!");
		return 0;
	}
	dumb_resampling_quality = 1;
	atexit(&dumb_exit);
	dumb_register_packfiles();
	
	train_pic = load_pcx("train.pcx", game_palette);
	game_palette[0].r = 0;
	game_palette[0].g = 0;
	game_palette[0].b = 0;
	
	/* create truck */
	truck.image = load_pcx("truck.pcx", NULL);
	truck.mask_image = create_bitmap(24, 24);
	clear_to_color(truck.mask_image, 1);
	truck.mask = create_allegro_pmask(truck.mask_image);
	
	level_pic = load_pcx("level1.pcx", NULL);
	level_mask_pic = load_pcx("level1m.pcx", NULL);
	level_mask = create_allegro_pmask(level_mask_pic);
	piece_pic[0] = load_pcx("piece1.pcx", NULL);
	piece_pic[1] = load_pcx("piece2.pcx", NULL);
	piece_pic[2] = load_pcx("piece3.pcx", NULL);
	piece_pic[3] = load_pcx("piece4.pcx", NULL);
	piece_pic[4] = load_pcx("piece5.pcx", NULL);
	piece_pic[5] = load_pcx("piece6.pcx", NULL);
	piece_pic[6] = load_pcx("piece7.pcx", NULL);
	piece_pic[7] = load_pcx("piece8.pcx", NULL);
	piece_pic[8] = load_pcx("piece9.pcx", NULL);
	track_pic = load_pcx("track.pcx", NULL);
	game_over_pic = load_pcx("gameover.pcx", NULL);
	set_palette(game_palette);
	install_ticker(60);
	return 1;
}

int check_level_collision(void)
{
	if(check_pmask_collision(level_mask, truck.mask, 0,0, fixtoi(truck.x) + 20, fixtoi(truck.y) + 4))
	{
		return 1;
	}
	return 0;
}

void generate_piece(void)
{
	switch(truck.piece)
	{
		case 0:
		{
			piece[0][0] = 0;
			piece[0][1] = 1;
			piece[0][2] = 0;
			piece[0][3] = 0;
			piece[1][0] = 1;
			piece[1][1] = 1;
			piece[1][2] = 1;
			piece[1][3] = 0;
			piece[2][0] = 0;
			piece[2][1] = 1;
			piece[2][2] = 0;
			piece[2][3] = 0;
			piece[3][0] = 0;
			piece[3][1] = 0;
			piece[3][2] = 0;
			piece[3][3] = 0;
			break;
		}
	}
}

int place_piece(int x, int y)
{
	int i, j;
	
	for(i = 0; i < 4; i++)
	{
		for(j = 0; j < 4; j++)
		{
			if(piece[i][j] && board[(y - 16 + i * 16) / 16][(x - 288 + j * 16) / 16])
			{
				return 0;
			}
		}
	}
	for(i = 0; i < 4; i++)
	{
		for(j = 0; j < 4; j++)
		{
			if(piece[i][j])
			{
				board[(y - 16 + i * 16) / 16][(x - 288 + j * 16) / 16] = 1;
			}
		}
	}
	return 1;
}

void position_piece(BITMAP * bp, int x, int y)
{
	int i, j;
	
	for(i = 0; i < 4; i++)
	{
		for(j = 0; j < 4; j++)
		{
			if(piece[i][j])
			{
				putpixel(bp, ((x - 288) / 16) * 16 + 8 + 288 + j * 16, ((y - 16) / 16) * 16 + 8 + 16 + i * 16, makecol(0, 0, 0));
			}
//			if(piece[i][j] && board[(mouse_y - 16) / 16][(mouse_x - 288) / 16])
//			{
//				return 0;
//			}
		}
	}
}

void tins06_logic(void)
{
	if(key[KEY_ESC])
	{
		user_exit = 1;
	}
	switch(state)
	{
		case 0:
		{
			if(-640 + train_pos / 15 >= 0)
			{
				game_state = 2;
			}
			switch(game_state)
			{
				case 0:
				{
					if(key[KEY_S])
					{
						train_pos += 30;
					}
					train_pos++;
					if(key[KEY_UP])
					{
						truck.x = fadd(truck.x, fcos(itofix(truck.angle)));
						truck.y = fadd(truck.y, fsin(itofix(truck.angle)));
						if(check_level_collision())
						{
							truck.x = fsub(truck.x, fcos(itofix(truck.angle)));
							truck.y = fsub(truck.y, fsin(itofix(truck.angle)));
						}
					}
					if(key[KEY_DOWN])
					{
						truck.x = fsub(truck.x, fcos(itofix(truck.angle)));
						truck.y = fsub(truck.y, fsin(itofix(truck.angle)));
						if(check_level_collision())
						{
							truck.x = fadd(truck.x, fcos(itofix(truck.angle)));
							truck.y = fadd(truck.y, fsin(itofix(truck.angle)));
						}
					}
					if(key[KEY_LEFT])
					{
						truck.angle--;
						if(check_level_collision())
						{
							truck.angle++;
						}
					}
					if(key[KEY_RIGHT])
					{
						truck.angle++;
						if(check_level_collision())
						{
							truck.angle--;
						}
					}
					if(fixtoi(truck.x) > 800 && fixtoi(truck.y) < 40 && truck.piece >= 0)
					{
						game_state = 1;
						level_y = -128;
						generate_piece();
					}
					break;
				}
				case 1:
				{
					train_pos++;
					if(mouse_b & 1)
					{
						if(place_piece(mouse_x, mouse_y))
						{
							truck.piece = -1;
							game_state = 0;
//							level_y = 0;
						}
					}
					break;
				}
				case 2:
				{
					train_pos += 30;
					if(train_pos / 15 >= 1280)
					{
						game_state = 3;
					}
					break;
				}
				case 3:
				{
					user_exit = 1;
					break;
				}
			}
			break;
		}
	}
}

void tins06_render(void)
{
	int i, j;
	
	switch(state)
	{
		case 0:
		{
			switch(game_state)
			{
				case 3:
				{
					draw_ascii(back_buffer, game_over_pic);
					break;
				}
				default:
				{
					clear_bitmap(top_buffer);
					draw_sprite(top_buffer, train_pic, -640 + train_pos / 15, 64);
					if(fixtoi(truck.x) < 320)
					{
						level_x = 0;
					}
					else if(fixtoi(truck.x) >= level_pic->w - 320)
					{
		//				level_x = level_pic->w - 320;
					}
					else
					{
						level_x = 320 - fixtoi(truck.x);
					}
					rectfill(bottom_buffer, 0, 0, 335, 239, makecol(0, 192, 0));
					rectfill(bottom_buffer, 336, 0, 639, 239, makecol(128, 64, 64));
					blit(level_pic, bottom_buffer, 0, 0, level_x, -level_y, level_pic->w, level_pic->h);
					for(i = 0; i < 6; i++)
					{
						for(j = 0; j < 20; j++)
						{
							if(board[i][j])
							{
								rectfill(bottom_buffer, level_pic->w - (j * 16) + level_x, -128 + i * 16 + 16 - level_y, level_pic->w - (j * 16) + level_x + 15, -128 + i * 16 + 16 - level_y + 15, makecol(96, 96, 96));
							}
						}
					}
					for(i = 0; i < (level_pic->w - 288) / 32; i++)
					{
						draw_sprite(bottom_buffer, track_pic, i * 32 + level_x, -128 + 32 - level_y);
					}
					rotate_sprite(bottom_buffer, truck.image, fixtoi(truck.x) + level_x, fixtoi(truck.y) - level_y, itofix(truck.angle));
					if(game_state == 0 && truck.piece >= 0)
					{
						pivot_sprite(bottom_buffer, piece_pic[truck.piece], fixtoi(truck.x) + level_x + 32, fixtoi(truck.y) - level_y + 16, 32, 16, itofix(truck.angle));
					}
		//			draw_sprite(bottom_buffer, truck_buffer, fixtoi(truck_x) + level_x, fixtoi(truck_y) - level_y);
					for(i = 0; i < 5; i++)
					{
						draw_sprite(bottom_buffer, piece_pic[(int)piece_slot[i]], 32 + level_x, 50 - level_y + i * 24);
					}
					if(game_state == 1)
					{
						for(i = 0; i < 4; i++)
						{
							for(j = 0; j < 4; j++)
							{
								if(piece[i][j])
								{
									rectfill(bottom_buffer, mouse_x + j * 16, mouse_y + i * 16, mouse_x + j * 16 + 15, mouse_y + i * 16 + 15, makecol(96, 96, 96));
								}
							}
						}
					}
					break;
				}
			}
			break;
		}
	}
//	textprintf_ex(back_buffer, font, 0, 0, makecol(255, 255, 255), makecol(0, 0, 0), "%d", game_state);
	position_piece(bottom_buffer, mouse_x, mouse_y);
	blit(back_buffer, screen, 0, 0, 0, 0, back_buffer->w, back_buffer->h);
}

int main(int argc, char * argv[])
{
	if(!tins06_initialize())
	{
		return 0;
	}
	truck.x = itofix(50);
	truck.y = itofix(100);
	truck.angle = 0;
	train_pos = 0;
	generate_board();
	if(!play_xm("bgm.xm"))
	{
		allegro_message("Could not play music!");
		return 0;
	}
	reset_ticker();
	while(!user_exit)
	{
		while(ticker > 0)
		{
			tins06_logic();
			ticker--;
		}
		tins06_render();
		update_xm();
		rest(0);
	}
	stop_xm();
	return 0;
}
END_OF_MAIN()
